<?php

declare(strict_types=1);

namespace Drupal\editor_button_link\Plugin\CKEditor5Plugin;

use Drupal\ckeditor5\HTMLRestrictions;
use Drupal\ckeditor5\Plugin\CKEditor5PluginConfigurableTrait;
use Drupal\ckeditor5\Plugin\CKEditor5PluginDefault;
use Drupal\ckeditor5\Plugin\CKEditor5PluginConfigurableInterface;
use Drupal\ckeditor5\Plugin\CKEditor5PluginElementsSubsetInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\editor\EditorInterface;

/**
 * CKEditor 5 ButtonLink plugin configuration.
 *
 * @internal
 *   Plugin classes are internal.
 */
class ButtonLink extends CKEditor5PluginDefault implements CKEditor5PluginConfigurableInterface, CKEditor5PluginElementsSubsetInterface {

  use CKEditor5PluginConfigurableTrait;

  /**
   * The default configuration for this plugin.
   *
   * @var string[][]
   */
  const DEFAULT_CONFIGURATION = [
    'button_link_types' => [],
    'button_link_sizes' => [],
    'button_link_styles' => [],
  ];

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['button_link_types'] = [
      '#title' => $this->t('Button type'),
      '#type' => 'textarea',
      '#description' => $this->t('The possible values this option can contain. Enter one value per line, in the format key|label.'),
    ];
    $form['button_link_sizes'] = [
      '#title' => $this->t('Button size'),
      '#type' => 'textarea',
      '#description' => $this->t('The possible values this option can contain. Enter one value per line, in the format key|label.'),
    ];
    $form['button_link_styles'] = [
      '#title' => $this->t('Button style'),
      '#type' => 'textarea',
      '#description' => $this->t('The possible values this option can contain. Enter one value per line, in the format key|label.'),
    ];
    if (!empty($this->configuration['button_link_types'])) {
      $form['button_link_types']['#default_value'] = '';
      foreach ($this->configuration['button_link_types'] as $style) {
        $form['button_link_types']['#default_value'] .= $style['class'] . '|' . $style['label'] . PHP_EOL;
      }
    }
    if (!empty($this->configuration['button_link_sizes'])) {
      $form['button_link_sizes']['#default_value'] = '';
      foreach ($this->configuration['button_link_sizes'] as $style) {
        $form['button_link_sizes']['#default_value'] .= $style['class'] . '|' . $style['label'] . PHP_EOL;
      }
    }
    if (!empty($this->configuration['button_link_styles'])) {
      $form['button_link_styles']['#default_value'] = '';
      foreach ($this->configuration['button_link_styles'] as $style) {
        $form['button_link_styles']['#default_value'] .= $style['class'] . '|' . $style['label'] . PHP_EOL;
      }
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    // Match the config schema structure at ckeditor5.plugin.ckeditor5_style.
    $form_values = $form_state->getValues();
    foreach ($form_values as $form_name => $form_value){
      [$button_link_styles, $unparseable_lines] = self::parseStylesFormValue($form_value);
      if (!empty($unparseable_lines)) {
        $line_numbers = array_keys($unparseable_lines);
        $form_state->setError($form[$form_name], $this->formatPlural(
          count($unparseable_lines),
          'Line @line-number does not contain a valid value. Enter a valid CSS selector containing one or more classes, followed by a pipe symbol and a label.',
          'Lines @line-numbers do not contain a valid value. Enter a valid CSS selector containing one or more classes, followed by a pipe symbol and a label.',
          [
            '@line-number' => reset($line_numbers),
            '@line-numbers' => implode(', ', $line_numbers),
          ]
        ));
      }
      $form_state->setValue($form_name, $button_link_styles);
    } 
  }

  /**
   * Parses the line-based (for form) ButtonLink configuration.
   *
   * @param string $form_value
   *   A string containing >=1 lines with on each line a CSS selector targeting
   *   1 tag with >=1 classes, a pipe symbol and a label. An example of a single
   *   line: `p.foo.bar|Foo bar paragraph`.
   *
   * @return array
   *   The parsed equivalent: a list of arrays with each containing:
   *   - label: the label after the pipe symbol, with whitespace trimmed
   *   - element: the CKEditor 5 element equivalent of the tag + classes
   *
   * @internal
   *   This method is public only to allow the CKEditor 4 to 5 upgrade path to
   *   reuse this logic. Mark this private in https://www.drupal.org/i/3239012.
   *
   * @see \Drupal\ckeditor5\Plugin\CKEditor4To5Upgrade\Core::mapCKEditor4SettingsToCKEditor5Configuration()
   */
  public static function parseStylesFormValue(string $form_value): array {
    $unparseable_lines = [];

    $lines = explode("\n", $form_value);
    $button_link_styles = [];
    foreach ($lines as $index => $line) {
      if (empty(trim($line))) {
        continue;
      }

      // Validate the selector.
      $selector_matches = [];
      // @see https://www.w3.org/TR/CSS2/syndata.html#:~:text=In%20CSS%2C%20identifiers%20(including%20element,hyphen%20followed%20by%20a%20digit
      if (!preg_match('/(\.-?[_a-zA-Z]+[_a-zA-Z0-9-]*\s*)\|{1}([_a-zA-Z0-9-_\(\)\s]+)/', trim($line), $selector_matches)) {
        $unparseable_lines[$index + 1] = $line;
      } else {
        $button_link_styles[] = [
          'label' => $selector_matches[2],
          'class' => $selector_matches[1],
        ];
      }
    }
    return [$button_link_styles, $unparseable_lines];
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->configuration['button_link_types'] = $form_state->getValue('button_link_types');
    $this->configuration['button_link_sizes'] = $form_state->getValue('button_link_sizes');
    $this->configuration['button_link_styles'] = $form_state->getValue('button_link_styles');
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return static::DEFAULT_CONFIGURATION;
  }

  /**
   * {@inheritdoc}
   */
  public function getElementsSubset(): array {
    $subset = [];
    $subset[] = array_column($this->configuration['button_link_types'], 'element');
    $subset[] = array_column($this->configuration['button_link_sizes'], 'element');
    $subset[] = array_column($this->configuration['button_link_styles'], 'element');

    return $subset;
  }

  /**
   * {@inheritdoc}
   */
  public function getDynamicPluginConfig(array $static_plugin_config, EditorInterface $editor): array {
    $definitions = [];
    foreach ($this->configuration['button_link_types'] as $style) {
      // Transform configured button_link_styles to the configuration structure expected by
      // the CKEditor 5 ButtonLink plugin.
      $definitions[] = [
        'name' => $style['label'],
        'class' => $style['class'],
      ];
    }
    return [
      'style' => [
        'definitions' => $definitions,
      ],
    ];
  }

}
